'use strict';

const _ = require('lodash');
const util = require('util');

/* -------------------------------------------------------------------------
 * Base
 * ---------------------------------------------------------------------- */

class TypeBase {
    getName() {
        return this.constructor.name;
    }

    is(type) {
        return this.toString() == (type + '').toUpperCase();
    }

    isFamily(type) {
        return this.is(type);
    }

    isOneOf(types) {
        return types.map(t => t.toUpperCase()).includes(this.toString());
    }

    toString() {
        return this.getName().toUpperCase();
    }
}

/* -------------------------------------------------------------------------
 * BASIC
 * ---------------------------------------------------------------------- */

class STRING extends TypeBase {
    constructor(length) {
        super();
        this.length = length;
    }

    isFamily(type) {
        type = (type + '').toUpperCase();

        return type.match(/^(STRING|VARCHAR|TEXT|TINYTEXT)/);
    }
}

class NUMBER extends TypeBase {
    isFamily(type) {
        type = (type + '').toUpperCase();

        return type.match(/^(INTEGER|NUMBER|FLOAT|DOUBLE|DECIMAL|BIGINT|TINYINT|MEDIUMINT|INT)/);
    }
}

class INTEGER extends TypeBase {
}

class FLOAT extends TypeBase {
    isFamily(type) {
        type = (type + '').toUpperCase();

        return type.match(/^(NUMBER|DOUBLE|FLOAT)/);
    }
}

class BOOLEAN extends TypeBase {
}

class DATE extends TypeBase {
}

class TIME extends TypeBase {
}

class DATEONLY extends TypeBase {
    getName() {
        return 'DATETIME';
    }
}

class DATETIME extends TypeBase {
}

class TIMESTAMP extends TypeBase {
}

/* -------------------------------------------------------------------------
 * SPECIAL
 * ---------------------------------------------------------------------- */

class VIRTUAL extends TypeBase {
}

class JSONTYPE extends TypeBase {
    getName() {
        return 'JSON';
    }
}

class ARRAY extends TypeBase {
    constructor(fields) {
        super();
        this.fields = fields || {};
    }
}

class OBJECT extends TypeBase {
    constructor(fields) {
        super();
        this.fields = fields || {};
    }
}

class KEY extends TypeBase {
}

/* -------------------------------------------------------------------------
 * Utils
 * ---------------------------------------------------------------------- */

let makeTypeFactory = (type) => {
    const factory = function() {
        let args = Object.values(arguments);
        return new type(...args);
    };

    factory.class = type;
    factory.default = factory();
    factory.is = factory.default.is;
    factory.isFamily = factory.default.isFamily;

    return factory;
};

/* -------------------------------------------------------------------------
 * Types
 * ---------------------------------------------------------------------- */

let dataTypes = {
    STRING: STRING,
    // CHAR: CHAR,
    // TEXT: TEXT,
    NUMBER: NUMBER,
    INTEGER: INTEGER,
    // BIGINT: BIGINT,
    FLOAT: FLOAT,
    TIME: TIME,
    TIMESTAMP: TIMESTAMP,
    DATE: DATE,
    DATETIME: DATETIME,
    DATEONLY: DATEONLY,
    BOOLEAN: BOOLEAN,
    // NOW: NOW,
    // BLOB: BLOB,
    // DECIMAL: DECIMAL,
    // NUMERIC: DECIMAL,
    // UUID: UUID,
    // UUIDV1: UUIDV1,
    // UUIDV4: UUIDV4,
    // HSTORE: HSTORE,
    JSON: JSONTYPE,
    // JSONB: JSONB,
    VIRTUAL: VIRTUAL,
    ARRAY: ARRAY,
    OBJECT: OBJECT,
    // NONE: VIRTUAL,
    // ENUM: ENUM,
    // RANGE: RANGE,
    // REAL: REAL,
    // DOUBLE: DOUBLE,
    // 'DOUBLE PRECISION': DOUBLE,
    // GEOMETRY: GEOMETRY,
    // GEOGRAPHY: GEOGRAPHY,
    KEY: KEY,
};

let natives = _.clone(dataTypes);

for(let key in dataTypes) {
    dataTypes[key] = makeTypeFactory(dataTypes[key]);
}

dataTypes.natives = natives;

// dataTypes.postgres = require('./dialects/postgres/data-types')(dataTypes);
// dataTypes.mysql = require('./dialects/mysql/data-types')(dataTypes);
// dataTypes.mariadb = dataTypes.mysql;
// dataTypes.sqlite = require('./dialects/sqlite/data-types')(dataTypes);
// dataTypes.mssql = require('./dialects/mssql/data-types')(dataTypes);

module.exports = dataTypes;