'use strict';

const _ = require('lodash');
const asyncp = require('paradox-asyncp');
const validator = require('validator');
const moment = require('moment');
const SchemaTypes = require('./schema-types');

class Schema {

    constructor(fields, data, options) {
        this._fields = fields;
        this._data = data || {};
        this._options = _.defaults(options, {
            validate: {},
            getterMethods: {},
            setterMethods: {},
        });

        this.normalizeFields();
        this.applyValues(this._data, true);
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    normalizeFields() {
        let fields = this._fields;

        for(let key in fields) {
            let field = fields[key];

            if(typeof field == 'function') {
                fields[key] = {
                    type: field(),
                };

                field = fields[key];
            }

            if(typeof field == 'object' && !('type' in field)) {
                let fType = fields[key];

                fields[key] = {
                    type: fType,
                };

                field = fields[key];
            }

            if(typeof field.type == 'function') {
                field.type = fields[key].type();
            }

            if(!('defaultValue' in field)) {
                field['defaultValue'] = null;
            }

            if(!('allowNull' in field)) {
                field['allowNull'] = false;
            }
        }

        for(let methodName in this._options.getterMethods) {
            this[methodName] = this._options.getterMethods[methodName];
        }

        for(let methodName in this._options.setterMethods) {
            this[methodName] = this._options.setterMethods[methodName];
        }

        return this;
    }

    applyValues(data, fullApply) {
        let fields = this._fields;

        for(let fieldName in fields) {
            // Extract the value
            if(fieldName in data) {
                this[fieldName] = this.formatValue(fields[fieldName], data[fieldName]);

            } else {
                if(fullApply) {
                    if('defaultValue' in fields[fieldName]) {
                        this[fieldName] = fields[fieldName].defaultValue;

                    } else {
                        this[fieldName] = null;
                    }
                }
            }
        }

        return this;
    }

    getDataValue(key) {
        return this[key];
    }

    setDataValue(key, value) {
        this[key] = value;
        return this;
    }

    set (fieldName, value) {
        if(fieldName in this._options.setterMethods) {
            this._options.setterMethods[fieldName].apply(this, [value]);
            return this;
        }

        let fields = this._fields;

        if(fieldName in fields) {
            let field = fields[fieldName];

            if('set' in field) {
                let res = field['set'].apply(this, [value]);

                if(typeof res !== 'undefined') {
                    this[fieldName] = res;
                }

            } else {
                this[fieldName] = value;
            }

        } else {
            throw new Error('field not in field definition');
        }

        return this;
    }

    get (fieldName) {
        if(fieldName in this._options.getterMethods) {
            return this._options.getterMethods[fieldName].apply(this);
        }

        let fields = this._fields;

        if(fieldName in fields) {
            let field = fields[fieldName];

            if('get' in field) {
                return field['get'].apply(this, [this[fieldName]]);
            }

            if(fieldName in this) {
                return this[fieldName];

            } else {
                if('defaultValue' in field) {
                    return field.defaultValue;

                } else {
                    return null;
                }
            }

        } else {
            throw new Error('field not in field definition');
        }
    }

    getValues(raw) {
        let fields = this._fields;
        let values = {};

        for(let fieldName in fields) {
            if(raw) {
                values[fieldName] = this[fieldName];

            } else {
                values[fieldName] = this.get(fieldName);
            }
        }

        return values;
    }

    formatValue(field, value) {
        if(field.format) {
            return field.format.apply(this, [value]);
        }

        if(value === null) {
            return null;
        }

        // TODO: transformar solo cuando se recupera?

        if(field.type instanceof SchemaTypes.BOOLEAN.class) {
            value = value === true || value === 'true' || value === '1' || value === 1;
        }

        if(field.type instanceof SchemaTypes.NUMBER.class) {
            if(typeof value != 'number') {
                value = Number(value);
            }
        }

        if(field.type instanceof SchemaTypes.FLOAT.class) {
            value = parseFloat(value);
        }

        if(field.type instanceof SchemaTypes.INTEGER.class) {
            value = parseInt(value);
        }

        if(field.type instanceof SchemaTypes.DATE.class) {
            if(typeof value == 'string' && value) {
                let __ = require('../index');

                try {
                    value = __.time.strtotime(value);

                } catch(err) {
                    // IGNORAR
                }
            }
        }

        return value;
    }

    validate() {
        let toValidate = [];

        for(let key in this._fields) {
            let field = this._fields[key];
            let value = this.get(key);

            // conditional
            if(field.condition) {
                let meetCondition = true;

                _.forEach(field.condition, (condValue, subKey) => {
                    let subValue = this.get(subKey);

                    if(condValue != subValue) {
                        meetCondition = false;
                    }
                });

                // not meet condition, ignore validation
                if(!meetCondition) {
                    continue;
                }
            }

            if(value === null) {
                if(field.allowNull) {
                    continue;

                } else {
                    toValidate.push({
                        field: key,
                        value,
                        key: 'notNull',
                        options: true
                    });
                }
            }

            let validations = field.validate || {};

            for(let vkey in validations) {
                toValidate.push({
                    field: key,
                    value,
                    key: vkey,
                    options: validations[vkey],
                });
            }
        }

        return asyncp
            .eachLimit(toValidate, 1, validation => {
                return this
                    .validateValue(validation.value, validation.key, validation.options)
                    .catch(err => {
                        err = new Error('Validation error for field "' + validation.field + '": ' + err.message);
                        return Promise.reject(err);
                    })
            })
            .then(() => {
                let toValidate = [];

                for(let key in this._options.validate) {
                    toValidate.push({
                        name: key,
                        callback: this._options.validate[key],
                    });
                }

                return asyncp
                    .eachLimit(toValidate, 1, validation => {
                        let res = null;

                        try {
                            res = validation.callback.apply(this);

                        } catch(err) {
                            return Promise.reject(err);
                        }

                        return Promise
                            .resolve(res)
                            .then(validated => {
                                if(typeof validated === 'undefined') {
                                    return true;
                                }

                                if(!validated) {
                                    return Promise.reject(new Error('model validator failed: ' + validation.name));
                                }
                            })
                    });
            });
    }

    validateValue(value, validation, options) {
        let params = [
            value
        ];

        let validationKey = validation;

        let custom = {
            is: (value, regex) => {
                if(_.isArray(regex)) {
                    regex = new RegExp(regex[0], regex.length > 1 ? regex[1] : '');
                }

                return (value + '').match(regex);
            },
            not: (value, regex) => {
                if(_.isArray(regex)) {
                    regex = new RegExp(regex[0], regex.length > 1 ? regex[1] : '');
                }

                return !(value + '').match(regex);
            },
            isNull: (value) => {
                return value === null;
            },
            notNull: (value) => {
                return value !== null;
            },
            notEmpty: (value) => {
                return (value + '').length > 0;
            },
            isIn: (value, set) => {
                return set.includes(value);
            },
            notIn: (value, set) => {
                return !set.includes(value);
            },
            notContains: (value, find) => {
                return (value + '').includes(find);
            },
            len: (value, len) => {
                let strLength = (value + '').length;

                if(len.length == 1) {
                    return strLength > len[0];

                } else {
                    return strLength >= len[0] && strLength <= len[1];
                }
            },
            isDate: (value) => {
                return moment(value).isValid();
            },
            max: (value, max) => {
                return value <= max;
            },
            min: (value, min) => {
                return value >= min;
            }
            // isIPv4: null,
            // isIPv6: null,
        };

        if(!_.isPlainObject(options)) {
            options = {
                args: options,
                msg: 'value doesn\'t match assertion of type "' + validationKey + '"',
            };

        }

        options.msg = options.msg || 'value doesn\'t match assertion of type "' + validationKey + '"';

        let res = null;

        if(typeof options.args == 'function') {
            res = options.args.apply(this, params);

        } else if(validation in custom) {
            params.push(options.args);
            res = custom[validation].apply(this, params);

        } else {
            if(!(validation in validator)) {
                throw new Error('unknown validation: ' + validation);
            }

            params.push(options.args);
            res = validator[validation].apply(this, params);
        }

        return Promise
            .resolve(res)
            .then(validated => {
                if(typeof validated === 'undefined') {
                    return true;
                }

                if(!validated) {
                    return Promise.reject(new Error(options.msg));
                }
            });
    }
}

Schema.type = SchemaTypes;

module.exports = Schema;